/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *     Alexander Diewald (fortiss GmbH) - backport to AF3 & Integration
 *                                        improvements
 *
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;

import javafx.scene.control.TreeItem;

/**
 * {@link TreeItem} with support for dynamic children using the
 * {@link DynamicTreeItem#update()} method.
 */
public class DynamicTreeItem<T> extends DynamicTreeItemBase<T> {
	/** The viewer. */
	private DynamicTreeContentProviderBase<T> contentProvider;

	/** Constructor. */
	public DynamicTreeItem(T data, DynamicTreeContentProviderBase<T> viewer) {
		super(data);
		this.contentProvider = viewer;
	}

	/** {@inheritDoc} */
	@Override
	public void update() {

		// This method is written so that in the common cases of deleting a child, inserting a child
		// and swapping two children, it only updates the affected children. Especially in these
		// common cases, this is much faster than clearing and re-populating the list of children.
		List<TreeItem<T>> children = getChildren();

		HashMap<T, Boolean> expanded = new HashMap<>();
		for(TreeItem<T> child : children) {
			expanded.put(child.getValue(), child.isExpanded());
		}

		Collection<? extends T> updatedContent =
				contentProvider.getFilteredSortedChildren(getValue());

		// List of values that are new since the last update
		List<T> newlyAddedValues = new ArrayList<>(updatedContent);
		for(TreeItem<T> child : children) {
			newlyAddedValues.remove(child.getValue());
		}

		// Remove:
		children.removeIf(ti -> !updatedContent.contains(ti.getValue()));

		// Insert and Update:
		// In the common cases where only new values have been inserted or values have been
		// permuted, this will insert/modify only the affected values. This is faster than deleting
		// all children and adding them from scratch.
		int i = 0;
		for(T updatedValue : updatedContent) {
			DynamicTreeItem<T> child = null;
			if(i < children.size()) {
				if(updatedValue.equals(children.get(i).getValue())) {
					// Child at position i already has the right value
					child = (DynamicTreeItem<T>)children.get(i);
				} else {
					// There is a child at position i, but its value is incorrect
					child = new DynamicTreeItem<T>(updatedValue, contentProvider);
					child.setExpanded(expanded.getOrDefault(updatedValue, false));

					if(newlyAddedValues.contains(updatedValue)) {
						children.add(i, child);
					} else {
						children.set(i, child);
					}
				}
			} else {
				// There is no child at position i
				child = new DynamicTreeItem<T>(updatedValue, contentProvider);
				child.setExpanded(expanded.getOrDefault(updatedValue, false));
				children.add(child);
			}
			child.update();
			i++;
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean isLeaf() {
		Collection<? extends Object> children =
				contentProvider.getFilteredSortedChildren(getValue());
		return children == null || children.isEmpty();
	}
}
