/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Stack;

import org.eclipse.emf.common.util.EList;
import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.base.ExitConnectorBase;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.ext.quality.data.MetricData;
import org.fortiss.tooling.ext.quality.data.MetricKey;
import org.fortiss.tooling.ext.quality.data.MetricTreeNode;

/**
 * Supplies methods to compute certain graph metrics.
 * 
 * @author groh
 */

// TODO (#4332) Restructure GraphMetricsProvider
public class GraphMetricsProvider {

	/**
	 * Creates a list of nodes contained in the graph centered around the provided element.
	 * It will contain all elements contained within the provided element as well as all elements
	 * connected to the provided element.
	 * 
	 * @param scopeElement
	 *            element defining the scope of the graph
	 * @param recursively
	 *            if true, everything is resolved to leaf elements
	 * @return node list
	 */
	private static Set<IHierarchicElement> getLocalGraphView(IHierarchicElement scopeElement,
			boolean recursively) {
		// This ensures that elements inside this element communicating with elements outside
		// this element are properly recognized
		Set<IHierarchicElement> graphNodes = new HashSet<>();
		if(recursively) {
			recursivelyGetLeafElements(scopeElement, graphNodes);
		} else {
			graphNodes.addAll(scopeElement.getContainedElements());
		}

		// Collect all elements to which currentElement has a outgoing connection
		for(ExitConnectorBase exitConnector : pickInstanceOf(ExitConnectorBase.class,
				scopeElement.getConnectors())) {
			for(IConnection exitConnection : exitConnector.getOutgoing()) {
				graphNodes.add(exitConnection.getTarget().getOwner());
			}
		}
		// Collect all elements to which currentElement has a incoming connection
		for(EntryConnectorBase entryConnector : pickInstanceOf(EntryConnectorBase.class,
				scopeElement.getConnectors())) {
			for(IConnection entryConnection : entryConnector.getIncoming()) {
				graphNodes.add(entryConnection.getSource().getOwner());
			}
		}
		return graphNodes;
	}

	/**
	 * Get all leafs inside the provided element.
	 * 
	 * @param element
	 *            the element containing leafs
	 * @param allElements
	 *            list where the leafs are saved
	 */
	private static void recursivelyGetLeafElements(IHierarchicElement element,
			Set<IHierarchicElement> allElements) {
		allElements.add(element);
		for(IHierarchicElement containedElement : element.getContainedElements()) {
			recursivelyGetLeafElements(containedElement, allElements);
		}
	}

	/**
	 * Follows outgoing connections until the connector has no further outgoing connections and then
	 * saves the owner of this connector to the provided list.
	 * 
	 * @param connector
	 *            starting point for search
	 * @param allElements
	 *            list where the owners are collected
	 */
	private static void recursivlyFollowOutgoingConnection(IConnector connector,
			Set<IHierarchicElement> allElements) {

		EList<IConnection> outgoingConnections = connector.getOutgoing();

		if(outgoingConnections.isEmpty()) {
			allElements.add(connector.getOwner());
			return;
		}

		for(IConnection outgoingConnection : outgoingConnections) {
			recursivlyFollowOutgoingConnection(outgoingConnection.getTarget(), allElements);
		}
	}

	/**
	 * Follows incoming connections until the connector has no further incoming connections and
	 * then saves the owner of this connector to the provided list.
	 * 
	 * @param connector
	 *            starting point for search
	 * @param allElements
	 *            list where the owners are collected
	 */
	private static void recursivlyFollowIncomingConnection(IConnector connector,
			Set<IHierarchicElement> allElements) {

		EList<IConnection> incomingConnections = connector.getIncoming();

		if(incomingConnections.isEmpty()) {
			allElements.add(connector.getOwner());
			return;
		}

		for(IConnection incomingConnection : incomingConnections) {
			recursivlyFollowIncomingConnection(incomingConnection.getSource(), allElements);
		}
	}

	/**
	 * Collects all neighbors regardless of the direction of connection.
	 * 
	 * @param element
	 *            of which the neighbors should be collected
	 * @return set of the neighbors
	 */
	private static Set<IHierarchicElement> getUndirectedNeighbors(IHierarchicElement element) {
		Set<IHierarchicElement> neighbors = new HashSet<>();

		for(ExitConnectorBase exitConnectors : pickInstanceOf(ExitConnectorBase.class,
				element.getConnectors())) {
			recursivlyFollowOutgoingConnection(exitConnectors, neighbors);
		}
		for(EntryConnectorBase entryConnectors : pickInstanceOf(EntryConnectorBase.class,
				element.getConnectors())) {
			recursivlyFollowIncomingConnection(entryConnectors, neighbors);
		}
		return neighbors;
	}

	/**
	 * Calculates the clustering coefficient based on the neighborhood.
	 * See, e.g., https://en.wikipedia.org/wiki/Clustering_coefficient
	 * Clustering coefficient C for a node V is counting the triangles divided by the "number of
	 * neighbors of "V" times "number of neighbors of V - 1" which results in the potential number
	 * of maximum links between the nodes
	 * 
	 * @param element
	 *            of which the coefficient shall be calculated
	 * @return value of the coefficient
	 */
	public static double calculateClusteringCoefficent(IHierarchicElement element) {

		Set<IHierarchicElement> neighbors = getUndirectedNeighbors(element);
		int totalEdges = neighbors.size();
		if(totalEdges < 2) {
			return 0.0; // If there are fewer than 2 neighbors, the coefficient is undefined (or 0).
		}
		// Iterate over neighbors and get their neighbors and check if they are also our neighbor in
		// which case we have a triangle
		int triangles = 0;
		for(IHierarchicElement neighbor : neighbors) {
			Set<IHierarchicElement> neighborNeighbors = getUndirectedNeighbors(neighbor);
			for(IHierarchicElement secondNeighbor : neighbors) {
				if(secondNeighbor != neighbor && neighborNeighbors.contains(secondNeighbor)) {
					triangles++;
				}
			}
		}
		double clusteringCoefficient = (double)triangles / (totalEdges * (totalEdges - 1));
		return clusteringCoefficient;
	}

	/**
	 * Calculates and saves the betweenness centrality of all elements contained inside the provided
	 * scopeElement. See, e.g., https://en.wikipedia.org/wiki/Betweenness_centrality
	 * 
	 * @param scopeElement
	 *            the scope of this calculation
	 * @param metricData
	 *            location to save the metrics
	 * @param recursively
	 *            if true, all calculation will be carried out on leaf elements
	 *            if false, it is only computing the centrality with the children
	 *            which means the inner circle.
	 */
	public static void calculateBetweennessCentrality(IHierarchicElement scopeElement,
			MetricData metricData, boolean recursively) {

		// Abort if no elements are found
		if(scopeElement.getContainedElements().isEmpty()) {
			return;
		}

		Set<IHierarchicElement> graphNodes = getLocalGraphView(scopeElement, recursively);
		// Initialize the betweenness values for all nodes to zero
		Map<IHierarchicElement, Double> betweenness = new HashMap<>();
		for(IHierarchicElement elementNode : graphNodes) {
			betweenness.put(elementNode, 0.0);
		}

		// Iterate over all collected nodes
		for(IHierarchicElement elementNode : graphNodes) {

			// Initialize all variables with standard values
			Stack<IHierarchicElement> stack = new Stack<>();
			Map<IHierarchicElement, List<IHierarchicElement>> predecessors = new HashMap<>();
			Map<IHierarchicElement, Integer> distance = new HashMap<>();
			Map<IHierarchicElement, Double> dependency = new HashMap<>();
			Queue<IHierarchicElement> queue = new LinkedList<>();

			for(IHierarchicElement v : graphNodes) {
				predecessors.put(v, new ArrayList<>());
				distance.put(v, -1);
				dependency.put(v, 0.0);
			}
			dependency.put(elementNode, 1.0);
			distance.put(elementNode, 0);
			queue.add(elementNode);

			processCentrality(queue, stack, predecessors, distance, dependency, graphNodes,
					elementNode, recursively);

			// Initialize delta
			Map<IHierarchicElement, Double> delta = new HashMap<>();
			for(IHierarchicElement v : graphNodes) {
				delta.put(v, 0.0);
			}

			while(!stack.isEmpty()) {
				IHierarchicElement w = stack.pop();
				for(IHierarchicElement v : predecessors.get(w)) {
					double c = (dependency.get(v) / dependency.get(w)) * (1.0 + delta.get(w));
					delta.put(v, delta.get(v) + c);
				}
				if(!w.equals(elementNode)) {
					betweenness.put(w, betweenness.get(w) + delta.get(w));
				}
			}
		}
		// Save metric
		saveMetric(scopeElement, metricData, recursively, betweenness);
	}

	/** Saving the metrics for each contained element of an {@link IHierarchicElement}. */
	private static void saveMetric(IHierarchicElement scopeElement, MetricData metricData,
			boolean recursively, Map<IHierarchicElement, Double> betweenness) {
		MetricKey key = recursively ? MetricKey.BETWEENESS_CENTRALITY_RECURSIVELY
				: MetricKey.BETWEENESS_CENTRALITY;
		for(IHierarchicElement child : scopeElement.getContainedElements()) {
			MetricTreeNode node = metricData.getTreeNodeLookupTable().get(child);
			node.getDoubleMetrics().put(key, betweenness.get(child));
		}
	}

	/**
	 * Performs intermediate calculations for the betweenness centrality algorithm.
	 * 
	 * @param queue
	 *            of nodes for Breadth-First Search.
	 * @param stack
	 *            of nodes for Breadth-First Search.
	 * @param predecessor
	 *            maps to lists used to keep track of the predecessors of nodes in the shortest
	 *            paths.
	 * @param distance
	 *            map used to store the shortest distances from the source node to other nodes.
	 * @param dependency
	 *            map used to calculate and store the dependency values for nodes in the network.
	 *            These values are part of the calculation for betweenness centrality.
	 * @param graphNodes
	 *            list of nodes
	 * @param scopeElement
	 *            the scope of this calculation
	 * @param recursively
	 *            if true, all calculation will be carried out on leaf elements
	 */
	private static void processCentrality(Queue<IHierarchicElement> queue,
			Stack<IHierarchicElement> stack,
			Map<IHierarchicElement, List<IHierarchicElement>> predecessor,
			Map<IHierarchicElement, Integer> distance, Map<IHierarchicElement, Double> dependency,
			Set<IHierarchicElement> graphNodes, IHierarchicElement scopeElement,
			boolean recursively) {

		while(!queue.isEmpty()) {
			IHierarchicElement v = queue.remove();
			stack.push(v);

			// Iterate over all connectors going out from the element
			for(ExitConnectorBase exitConnectors : pickInstanceOf(ExitConnectorBase.class,
					v.getConnectors())) {
				// Iterate over all connections
				for(IConnection outgoingConnection : exitConnectors.getOutgoing()) {

					Set<IHierarchicElement> targetElements = new HashSet<>();
					if(recursively) {
						recursivlyFollowOutgoingConnection(outgoingConnection.getTarget(),
								targetElements);
					} else {
						IHierarchicElement primaryTargetElement =
								outgoingConnection.getTarget().getOwner();
						// When we have a connection to the scopeElement, we have a
						// connection leaving the local scope
						if(scopeElement == primaryTargetElement) {
							// Continue searching for outgoing connections on the target
							// Connector
							IConnector targetConnector = outgoingConnection.getTarget();
							for(IConnection secondaryOutgoingConnection : targetConnector
									.getOutgoing()) {
								targetElements
										.add(secondaryOutgoingConnection.getTarget().getOwner());
							}
						} else {
							targetElements.add(primaryTargetElement);
						}
					}
					for(IHierarchicElement targetElement : targetElements) {

						// Check if element is in node list, to avoid creating data for
						// nodes outside of our graph
						if(graphNodes.contains(targetElement)) {
							var targetDistance = distance.get(targetElement);
							var vDistance = distance.get(v);
							// first start of distance calculation from v to the target Element
							if(targetDistance < 0) {
								queue.add(targetElement);
								distance.put(targetElement, vDistance + 1);
							} else if(targetDistance == vDistance + 1) {
								dependency.put(targetElement,
										dependency.get(targetElement) + dependency.get(v));
								predecessor.get(targetElement).add(v);
							}
						}
					}
				}
			}
		}
	}
}
