/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;

/**
 * A general class for storing metric data in a hierarchical tree structure.
 * <p>
 * Allows for efficient storage and retrieval of data.
 * 
 * @author groh
 */
public class MetricTreeNode {

	/** stores the children of this tree node */
	private List<MetricTreeNode> children;

	/** Map containing all metrics which are a Integer */
	private Map<MetricKey, Integer> integerMetrics;

	/** Map containing all metrics which are a Double */
	private Map<MetricKey, Double> doubleMetrics;

	/** Map containing all metrics which are a String */
	private Map<MetricKey, String> stringMetrics;

	/** name of the element which is associated with this data */
	private String name;

	/** Constructor. */
	public MetricTreeNode() {
		children = new ArrayList<>();
		integerMetrics = new HashMap<>();
		doubleMetrics = new HashMap<>();
		stringMetrics = new HashMap<>();
	}

	/** Getter for {@link #children}. */
	public List<MetricTreeNode> getChildren() {
		return children;
	}

	/** Getter for {@link #name}. */
	public String getName() {
		return name;
	}

	/** Setter for {@link #name}. */
	public void setName(String name) {
		this.name = name;
	}

	/** Getter for {@link #integerMetrics}. */
	public Map<MetricKey, Integer> getIntegerMetrics() {
		return integerMetrics;
	}

	/** Getter for {@link #doubleMetrics}. */
	public Map<MetricKey, Double> getDoubleMetrics() {
		return doubleMetrics;
	}

	/** Getter for {@link #stringMetrics}. */
	public Map<MetricKey, String> getStringMetrics() {
		return stringMetrics;
	}

	/**
	 * Allows to retrieve allows to retrieve the value of a key which may be stored as double or
	 * integer type.
	 * In the latter case the value will automatically be converted into a double.
	 * 
	 * @param key
	 *            for which the value shall be retrieved
	 * @return the respective integer or double
	 */
	public Double getValueAsDouble(MetricKey key) {
		if(key.isDoubleValue()) {
			return doubleMetrics.get(key);
		}
		Integer val = integerMetrics.get(key);
		return val == null ? null : val.doubleValue();
	}

	/**
	 * Traverses the tree in post-order.
	 * 
	 * @param consumer
	 *            which is applied to each node
	 */
	public void traverseTree(Consumer<MetricTreeNode> consumer) {

		// traverse all children
		for(var child : getChildren()) {
			child.traverseTree(consumer);
		}

		// apply consumer to self
		consumer.accept(this);
	}
}
