/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.storage;

import static java.util.stream.Collectors.toList;
import static org.apache.commons.io.FilenameUtils.getBaseName;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.eclipse.core.runtime.NullProgressMonitor;
import org.fortiss.tooling.ext.quality.data.DataRootElement;
import org.fortiss.tooling.ext.quality.data.MetricKey;
import org.fortiss.tooling.ext.quality.service.IModelQualityService;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;

/**
 * Allows the export of metrics to a CSV file.
 * 
 * @author groh
 */
public class CSVFileWriter {

	/** Suffix of the file in which the data is written. */
	private static final String CSV = ".csv";

	/**
	 * Method to write the data into CSV files. For each project, one file is built. If it already
	 * exists, it uses to the existing keys in the file as dimensions, to help that the log file
	 * always remains consistent.
	 * 
	 * TODO (#4334) Newly added {@link MetricKey}s are not added in the next export
	 * 
	 * @param map
	 *            to save into CSV file
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 * 
	 */
	public static void metricExtractionToCSV(Map<IProjectRootElement, DataRootElement> map)
			throws NoSuchAlgorithmException, IOException {

		// check if there is any data to export
		if(map == null || map.isEmpty()) {
			return;
		}

		IModelQualityService.getInstance().createQualityFolder();

		// Map is not empty due to check above
		DataRootElement firstValue = map.entrySet().iterator().next().getValue();
		ITopLevelElement topLevelElement = firstValue.getTopLevelElement();
		String projectName = topLevelElement.getSaveableName();

		var fileBaseName = getBaseName(projectName).toString();
		// TODO (#4333)
		Path path = new File(ModelQualityStorageManager.MODEL_QUALITY_PROJECT_DIR,
				fileBaseName.concat(CSV)).toPath();

		// Metric extraction is triggered when model files are saved. Therefore, the topLevelElement
		// should never be dirty here. In order to be on the safe side, conditionally save model.
		// save before using the saved file to produce the hash
		if(topLevelElement.isDirty()) {
			try {
				topLevelElement.doSave(new NullProgressMonitor());
				// Avoid duplicated metric extraction (the save operation will trigger another one).
				return;
			} catch(Exception e) {
				e.printStackTrace();
			}
			return;
		}
		// TODO (#4333)
		Path modelFile = new File(ModelQualityStorageManager.MODEL_PROJECT_DIR,
				topLevelElement.getSaveableName()).toPath();
		String projectHash = computeGitObjectHash(modelFile);

		List<String> allKeys = new ArrayList<>();
		boolean createNewIndex = true;

		if(Files.exists(path)) {
			// Read first line and try to extract the keys from the already existing file
			try(BufferedReader reader = new BufferedReader(new FileReader(path.toFile()))) {
				String line = reader.readLine();
				if(line != null) {
					allKeys = Arrays.asList(line.split(","));
					createNewIndex = false;
				}
			} catch(IOException e) {
				// In case there is an error, writeMetricsToFile() will provide the default keys.
				// Therefore this exception can be ignored (and just logged).
				e.printStackTrace();
			}
		}
		writeMetricsToFile(map, allKeys, path, createNewIndex, projectName, projectHash);
	}

	/**
	 * Function to write the specified metrics into a CSV file.
	 * 
	 * @param map
	 *            of the metrics which are written into the file
	 * @param allKeys
	 *            a list of keys which will be extracted from the provided map
	 * @param path
	 *            location of the file
	 * @param createNewIndex
	 *            if a index should be written before writing any values
	 * @param projectName
	 *            the name of the base fileProject for this map currently in the focus
	 * @param projectHash
	 *            the model file hash value saved for version identification
	 */
	private static void writeMetricsToFile(Map<IProjectRootElement, DataRootElement> map,
			List<String> allKeys, Path path, boolean createNewIndex, String projectName,
			String projectHash) {

		try(var writer = new BufferedWriter(new FileWriter(path.toFile(), true))) {
			if(createNewIndex) {
				// Create new index and write it into the first line of the file.
				allKeys.add("nv_timestamp");
				allKeys.add("nv_project_name");
				allKeys.add("nv_project_hash");
				allKeys.add("nv_root_element_name");
				allKeys.add("nv_name");
				allKeys.add("nv_children");
				// Collect all MetricKeys, convert them to a string and append them to the allKeys
				// list
				allKeys.addAll(Arrays.stream(MetricKey.values()).map(MetricKey::toString)
						.map(String::toLowerCase).collect(toList()));

				// Write first line
				writer.write(String.join(",", allKeys));
				writer.write("\n");
			}

			// Remove first 6 keys as they are not provided by the provider
			List<String> valueKeys = allKeys.subList(6, allKeys.size());

			LocalDateTime dateTime = LocalDateTime.now();
			DateTimeFormatter formatter = DateTimeFormatter.ISO_LOCAL_DATE_TIME;
			String formattedDateTime = dateTime.format(formatter);

			map.forEach((rootElement, dataRootElement) -> {
				// iterate over all project roots
				String rootName = rootElement.getClass().getSimpleName();
				var rootNode = dataRootElement.getRootNode();
				rootNode.traverseTree(node -> {
					var integers = node.getIntegerMetrics();
					var doubles = node.getDoubleMetrics();
					var strings = node.getStringMetrics();
					// collect IDs of children
					String children_ids = node.getChildren().stream()
							.map(n -> n.getIntegerMetrics().get(MetricKey.UNIQUE_ID))
							.map(id -> id.toString()).reduce((a, b) -> a + " " + b).orElse("");

					// These 6 data entries are just concatenated in front
					var startStream = Stream.of(formattedDateTime, projectName, projectHash,
							rootName, node.getName(), children_ids);
					// Collect all values from provider in the correct order and combine
					var values = Stream.concat(startStream, valueKeys.stream()
							.map(String::toUpperCase).map(MetricKey::valueOf).map(k -> {
								if(k.isStringValue()) {
									return strings.get(k);
								} else if(k.isDoubleValue()) {
									return doubles.get(k);
								}
								return integers.get(k);
							}).map(String::valueOf)).collect(Collectors.joining(","));
					try {
						// write values
						writer.write(values);
						writer.write("\n");
					} catch(IOException e) {
						e.printStackTrace();
					}
				});
			});
		} catch(IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Return the Git object hash for the contents of the given file:
	 * <p>
	 * {@code SHA-1("blob <size>\0"+<contents>).asHexString()}.
	 */
	private static String computeGitObjectHash(Path filePath)
			throws NoSuchAlgorithmException, IOException {

		try(BufferedReader reader = new BufferedReader(new FileReader(filePath.toString()))) {
			// Read file line-wise with UNIX line endings
			StringBuilder inputBuilder = new StringBuilder();
			do {
				String line = reader.readLine();
				if(line == null) {
					break;
				}
				inputBuilder.append(line);
				inputBuilder.append("\n");
			} while(true);

			// Construct pre-image (input to hash function) according to Git specification
			String fileContents = inputBuilder.toString();
			// fileContexts.length() does not return the correct length on some platforms (e.g.,
			// Linux)
			int n = fileContents.getBytes().length;
			String preImage = "blob " + n + "\0" + fileContents;

			// Compute hash and convert it to a hex string
			MessageDigest sha1 = MessageDigest.getInstance("SHA-1");
			byte[] digest = sha1.digest(preImage.getBytes());

			StringBuilder resultBuilder = new StringBuilder();
			for(byte b : digest) {
				resultBuilder.append(String.format("%02x", b & 0xff));
			}
			return resultBuilder.toString();
		}
	}
}
