/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.service;

import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.service.base.IEObjectAware;

/**
 * Interface for ReuseProvider that can be registered in the
 * {@link IReuseProviderService}.
 *
 * @author bergemann
 *
 * @param <T>
 *            The subclass of {@link EObject} for which this provider shall be
 *            invoked.
 */
public interface IReuseProvider<T extends EObject> extends IEObjectAware<T> {

	/**
	 * Prepares the given <T> for its storage in a reuse library. This might e.g. include cleaning
	 * it from attributes that are not needed anymore when "exported" into a library.
	 * 
	 * @param element
	 *            The <T> that should be prepared.
	 * @return True if preparation was successful, otherwise false.
	 */
	public boolean prepareReuseElementForLibrary(T element);

	/**
	 * Updates the first given <T>, which is a reuse element, based on the second
	 * given <T>. The update is a replacement of the old element with the new one,
	 * while also all references in depending elements are updated/replaced.
	 *
	 * @param elementToBeUpdated
	 *            The <T> that should be updated/replaced.
	 * @param newElement
	 *            The <T> with which the first one should be replaced
	 *            (updated).
	 * @return True if update was successful, otherwise false
	 */
	public boolean updateReuseElement(T elementToBeUpdated, T newElement);

	/**
	 * Sets the name of the given <T> used for reuse with the given {@link String}.
	 *
	 * @param reuseElement
	 *            The <T> whose name should be set.
	 * @param name
	 *            The {@link String} for the name.
	 * @return True if update was successful, otherwise false
	 */
	public boolean setReuseElementName(T reuseElement, String name);

	/**
	 * Gets the name of the given <T> used for reuse.
	 *
	 * @param reuseElement
	 *            The <T> whose name is requested.
	 * @return The name as {@link String}.
	 */
	public String getReuseElementName(T reuseElement);

	/**
	 * Sets the comment of the given <T> used for reuse with the given
	 * {@link String}.
	 *
	 * @param reuseElement
	 *            The <T> whose name should be set.
	 * @param comment
	 *            The {@link String} for the comment.
	 * @return True if update was successful, otherwise false
	 */
	public boolean setReuseElementComment(T reuseElement, String comment);

	/**
	 * Gets the comment of the given <T>} used for reuse.
	 *
	 * @param reuseElement
	 *            The <T> whose name is requested.
	 * @return The comment as {@link String}.
	 */
	public String getReuseElementComment(T reuseElement);

	/**
	 * Gets a list of classes of {@link EObject}s that could be possible external
	 * references within a reuse element (e.g. an external function inside a
	 * component code specification).
	 *
	 * @return The list (empty list if nothing could be found).
	 */
	public List<Class<? extends EObject>> getPossibleExternalReferenceClasses();
}
