/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.service;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;

/**
 * Interface of the {@link ReuseProviderService} which provides all
 * element-specific services for reuse like update/replace a (reuse) element
 * with a new version of it.
 *
 * @author bergemann
 *
 */
public interface IReuseProviderService extends IIntrospectiveKernelService {

	/** Name of the respective extension point. */
	static final String EXTENSION_POINT_NAME = "org.fortiss.tooling.ext.reuse.reuseProvider";
	/** Name of the respective configuration element. */
	static final String CONFIGURATION_ELEMENT_NAME = "reuseProvider";

	/**
	 * Retrieves the singleton instance of {@link ReuseProviderService}.
	 *
	 * @return The singleton instance of this service.
	 */
	public static IReuseProviderService getInstance() {
		return ReuseProviderService.getInstance();
	}

	/**
	 * Registers the given {@link IReuseProvider} in this service.
	 *
	 * @param provider
	 *            The {@link IReuseProvider} to register.
	 * @param cls
	 *            The subclass of {@link EObject} the given provider shall be
	 *            used for.
	 */
	public <T extends EObject> void registerReuseProvider(IReuseProvider<T> provider, Class<T> cls);

	/**
	 * Prepares the given reuse element as {@link EObject} for storage in a reuse library. This
	 * might include e.g. cleaning the reuse element from all attributes that are not needed anymore
	 * when "exported" into a library.
	 * 
	 * @param reuseElement
	 *            The reuse element that should be prepared.
	 * @return True if preparation was successful, otherwise false.
	 */
	public boolean prepareReuseElementForLibrary(EObject reuseElement);

	/**
	 * Updates the first given {@link EObject}, which is a reuse element, based on
	 * the second given {@link EObject}. The update is a replacement of the old
	 * element with the new one, while also all references in depending elements are
	 * updated/replaced.
	 *
	 * @param elementToBeUpdated
	 *            The {@link EObject} that should be
	 *            updated/replaced.
	 * @param newElement
	 *            The {@link EObject} with which the first one should
	 *            be replaced (updated).
	 * @return True if update was successful, otherwise false.
	 */
	public boolean updateReuseElement(EObject elementToBeUpdated, EObject newElement);

	/**
	 * Sets the name of the given {@link EObject} used for reuse with the given
	 * {@link String}.
	 *
	 * @param reuseElement
	 *            The {@link EObject} whose name should be set.
	 * @param name
	 *            The {@link String} for the name.
	 * @return True if update was successful, otherwise false.
	 */
	public boolean setReuseElementName(EObject reuseElement, String name);

	/**
	 * Gets the name of the given {@link EObject} used for reuse.
	 *
	 * @param reuseElement
	 *            The {@link EObject} whose name is requested.
	 * @return The name as {@link String}.
	 */
	public String getReuseElementName(EObject reuseElement);

	/**
	 * Sets the comment of the given {@link EObject} used for reuse with the given
	 * {@link String}.
	 *
	 * @param reuseElement
	 *            The {@link EObject} whose name should be set.
	 * @param comment
	 *            The {@link String} for the comment.
	 * @return True if update was successful, otherwise false.
	 */
	public boolean setReuseElementComment(EObject reuseElement, String comment);

	/**
	 * Gets the comment of the given {@link EObject} used for reuse.
	 *
	 * @param reuseElement
	 *            The {@link EObject} whose name is requested.
	 * @return The comment as {@link String}.
	 */
	public String getReuseElementComment(EObject reuseElement);
}
