/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.service;

import static org.eclipse.emf.ecore.util.EcoreUtil.replace;
import static org.fortiss.tooling.ext.variability.util.VariabilityUtils.getOptVarPointSpecification;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.replaceEObjectReferences;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IHierarchicElementContainer;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.ext.variability.model.OptionalVariationPointSpecification;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.model.INamedElement;

/***
 * Implemented basic {@link IReuseProvider} based on {@link EObject}. Can be
 * extended or overridden by more specific implemented {@link IReuseProvider}.
 *
 * @author bergemann
 *
 * @param <T>
 *            The subclass of {@link EObject} for which this provider base shall
 *            be extended.
 */
public class ReuseProviderBase<T extends EObject> implements IReuseProvider<T> {

	/** {@inheritDoc} */
	@Override
	public boolean prepareReuseElementForLibrary(T reuseElement) {
		if(reuseElement instanceof IModelElement) {
			removeVariabilitySpecifications((IModelElement)reuseElement);
		}

		removeExternalReferences(reuseElement);

		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean updateReuseElement(T elementToBeUpdated, T newElement) {
		replaceEObjectReferences(elementToBeUpdated, newElement);
		replace(elementToBeUpdated, newElement);
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean setReuseElementName(T reuseElement, String name) {
		if(reuseElement instanceof INamedElement) {
			((INamedElement)reuseElement).setName(name);
			return true;
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public String getReuseElementName(T reuseElement) {
		if(reuseElement instanceof INamedElement) {
			return ((INamedElement)reuseElement).getName();
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean setReuseElementComment(T reuseElement, String comment) {
		if(reuseElement instanceof INamedCommentedElement) {
			((INamedCommentedElement)reuseElement).setComment(comment);
			return true;
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public String getReuseElementComment(T reuseElement) {
		if(reuseElement instanceof INamedCommentedElement) {
			return ((INamedCommentedElement)reuseElement).getComment();
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public List<Class<? extends EObject>> getPossibleExternalReferenceClasses() {
		// Can only be specified for specific elements (if wanted) --> here only empty.
		return new ArrayList<Class<? extends EObject>>();
	}

	/**
	 * Removes unnecessary variability specifications from the given element and from all of its
	 * contained elements. The method is recursion-based!
	 */
	private void removeVariabilitySpecifications(IModelElement nextElement) {
		// Removal.
		OptionalVariationPointSpecification unnecessarySpec =
				getOptVarPointSpecification(nextElement);
		if(unnecessarySpec != null) {
			nextElement.getSpecifications().remove(unnecessarySpec);
		}

		// Recursion.
		if(nextElement instanceof IHierarchicElementContainer) {
			for(IHierarchicElement elem : ((IHierarchicElementContainer)nextElement)
					.getContainedElements()) {
				removeVariabilitySpecifications(elem);
			}
		}
	}

	/**
	 * Recursively removes all external (i.e., non-containment) reference targets from this and all
	 * contained {@link EObject}s.
	 * 
	 * FIXME #4150 (https://git.fortiss.org/af3/af3/-/issues/4150): This functionality should be
	 * removed, as soon as there is a dedicated management of external references.
	 * 
	 * @param element
	 *            The {@link EObject} for which all reference targets shall be removed.
	 */
	private void removeExternalReferences(EObject element) {
		// Whitelist all contained elements as reference targets.
		Set<EObject> whitelist = new HashSet<EObject>(getChildrenWithType(element, EObject.class));

		removeReferences(element, whitelist);
	}

	/**
	 * Recursively removes all reference targets from this and all contained {@link EObject}s which
	 * are not contained in the given whitelist.
	 * 
	 * FIXME #4150 (https://git.fortiss.org/af3/af3/-/issues/4150): This functionality should be
	 * removed, as soon as there is a dedicated management of external references.
	 */
	@SuppressWarnings("unchecked")
	private void removeReferences(EObject element, Set<EObject> whitelist) {
		// FIXME (SeB): the loop below does not get all internal/nested references, which is why it
		// will skip the removals of internal references due to no direct references in the given
		// element.
		for(EReference ref : element.eClass().getEReferences()) {
			Object refTarget = element.eGet(ref);

			if(refTarget == null) {
				continue;
			}

			if(refTarget instanceof Collection<?>) {
				Collection<EObject> refTargetCollection = (Collection<EObject>)refTarget;

				if(ref.isContainment()) {
					// Containment relations must not be removed. Yet, references of children need
					// to be removed recursively.
					refTargetCollection.stream()
							.forEach(target -> removeReferences(target, whitelist));
				} else {
					refTargetCollection.removeIf(target -> !whitelist.contains(target));
				}
			} else {
				if(ref.isContainment()) {
					// Keep reference target and make recursive call.
					removeReferences((EObject)refTarget, whitelist);
				} else {
					if(!whitelist.contains(refTarget)) {
						element.eSet(ref, null);
					}
				}
			}
		}
	}
}
