/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.service;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;

/**
 * Implementation of the {@link IReuseProviderService} that provides all
 * element-specific services for reuse like update/replace a (reuse) element
 * with a new version of it.
 *
 * @author bergemann
 */
public class ReuseProviderService extends EObjectAwareServiceBase<IReuseProvider<EObject>>
		implements IReuseProviderService {

	/** Singleton instance. */
	private static final ReuseProviderService INSTANCE = new ReuseProviderService();

	/**
	 * List of reuse element {@link Class}es ({@link EObject} {@link Class}es that
	 * provide reuse functionalities).
	 */
	private static List<Class<? extends EObject>> reuseElementClasses = new ArrayList<>();

	/**
	 * List of {@link EObject} {@link Class}es that might exist inside reuse
	 * elements as external references.
	 */
	private static List<Class<? extends EObject>> possibleExternalReferenceClasses =
			new ArrayList<>();

	/**
	 * Returns the singleton instance of {@link ReuseProviderService}.
	 *
	 * @return The singleton instance of this service
	 */
	public static ReuseProviderService getInstance() {
		return INSTANCE;
	}

	/**
	 * Returns all {@link Class}es that were registered at this service as reuse
	 * element.
	 *
	 * @return List of all registered reuse element classes
	 */
	public static List<Class<? extends EObject>> getReuseElementClasses() {
		return reuseElementClasses;
	}

	/**
	 * Returns all {@link Class}es that were registered at this service as possible
	 * external reference elements that could reference reuse elements.
	 *
	 * @return List of all registered external reference element classes for reuse
	 */
	public static List<Class<? extends EObject>> getPossibleExternalReferenceClasses() {
		return possibleExternalReferenceClasses;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <T extends EObject> void registerReuseProvider(IReuseProvider<T> provider,
			Class<T> cls) {
		addHandler(cls, (IReuseProvider<EObject>)provider);
		if(!reuseElementClasses.contains(cls)) {
			reuseElementClasses.add(cls);
		}
		possibleExternalReferenceClasses.addAll(provider.getPossibleExternalReferenceClasses());
	}

	/** {@inheritDoc} */
	@Override
	public boolean prepareReuseElementForLibrary(EObject reuseElement) {
		IReuseProvider<EObject> provider = getReuseProvider(reuseElement);
		if(provider != null) {
			return provider.prepareReuseElementForLibrary(reuseElement);
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean updateReuseElement(EObject elemToBeUpdated, EObject newElem) {
		IReuseProvider<EObject> provider = getReuseProvider(elemToBeUpdated);
		if(provider != null) {
			return provider.updateReuseElement(elemToBeUpdated, newElem);
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean setReuseElementName(EObject reuseElement, String name) {
		IReuseProvider<EObject> provider = getReuseProvider(reuseElement);
		if(provider != null) {
			return provider.setReuseElementName(reuseElement, name);
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public String getReuseElementName(EObject reuseElement) {
		IReuseProvider<EObject> provider = getReuseProvider(reuseElement);
		if(provider != null) {
			return provider.getReuseElementName(reuseElement);
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean setReuseElementComment(EObject reuseElement, String comment) {
		IReuseProvider<EObject> provider = getReuseProvider(reuseElement);
		if(provider != null) {
			return provider.setReuseElementComment(reuseElement, comment);
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public String getReuseElementComment(EObject reuseElement) {
		IReuseProvider<EObject> provider = getReuseProvider(reuseElement);
		if(provider != null) {
			return provider.getReuseElementComment(reuseElement);
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Reuse Provider Service";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return "Service providing reuse methods for given (reuse) elements.";
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return "provider";
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/**
	 * Returns the reuse provider/handler for the class of the given element. If none can be found,
	 * null is returned.
	 */
	private IReuseProvider<EObject> getReuseProvider(EObject element) {
		List<IReuseProvider<EObject>> registeredHandlers =
				getRegisteredHandlers(element.getClass());
		// There should only be one handler registered per class for the reuse case.
		if(registeredHandlers != null && registeredHandlers.size() == 1) {
			// get(0) is safe due to size check above.
			return registeredHandlers.get(0);
		}
		return null;
	}
}
