package org.fortiss.variability.analysis;

import static com.microsoft.z3.Status.SATISFIABLE;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.Expr;
import com.microsoft.z3.Model;
import com.microsoft.z3.Solver;
import com.microsoft.z3.Status;

/**
 * Base class that implements the translation of model product-lines to SMT in order to perform
 * product-line analysis via constraint lifting.
 * 
 * @author bayha
 */
public abstract class GenericProductLineAnalysis {

	/** The product-line translation to be used. */
	protected final IProductLineTranslation translation;

	/** {@link EObject} of the model to be translated. */
	protected final EObject model;

	/** The constraints to be checked by this analysis. */
	protected Collection<IProductLineConstraint> plConstraints;

	/**
	 * Constructor.
	 */
	protected GenericProductLineAnalysis(EObject model, IProductLineTranslation translation) {
		this.model = model;
		this.translation = translation;
	}

	/**  Creates the {@link IProductLineConstraint}s to be checked by this analysis. */
	protected abstract Collection<IProductLineConstraint> createConstraints();

	/**
	 * Performs a product-line analysis for the given model.
	 * 
	 * @return 'true' if all constraints are fulfilled for all variants of the product-line
	 *         analysis. 'false' otherwise.
	 */
	public List<ProductLineConstraintViolation> doCheck() {
		plConstraints = createConstraints();

		setTranslatedMetamodelElements();

		translation.translateModel(model);
		Map<BoolExpr, IProductLineConstraint> constraintTracker2Constraint =
				translation.translateConstraints(plConstraints);
		Solver solver = translation.createSolver();

		Status result = solver.check();

		List<ProductLineConstraintViolation> ret = new ArrayList<ProductLineConstraintViolation>();

		if(result.equals(SATISFIABLE)) {
			// A violating configuration (i.e., model for negated constraints) was found -> extract
			// violating model elements.
			Model z3Model = solver.getModel();

			for(BoolExpr constTrack : constraintTracker2Constraint.keySet()) {
				Expr<?> constInterp = z3Model.getConstInterp(constTrack);

				if(constInterp instanceof BoolExpr) {
					if(constInterp.equals(translation.getContext().mkTrue())) {
						IProductLineConstraint failedConstraint =
								constraintTracker2Constraint.get(constTrack);

						List<EObject> violatingObjects = new ArrayList<EObject>();
						for(Expr<?> varTrack : failedConstraint.getVariableTracker()) {
							Expr<?> interpr = z3Model.getConstInterp(varTrack);
							EObject eObject = translation.getEObjectForExpression(interpr);

							violatingObjects.add(eObject);
						}

						ret.add(new ProductLineConstraintViolation(failedConstraint,
								violatingObjects, null));
					}
				}

			}
		}

		translation.getContext().close();

		return ret;
	}

	/** Retrieves the classes, references and attributes required by the constraints. */
	private void setTranslatedMetamodelElements() {
		Set<EClass> classes = new HashSet<EClass>();
		Set<EReference> references = new HashSet<EReference>();
		Set<EAttribute> attributes = new HashSet<EAttribute>();

		for(IProductLineConstraint constraint : plConstraints) {
			classes.addAll(constraint.getTranslatedClasses());
			references.addAll(constraint.getTranslatedReferences());
			attributes.addAll(constraint.getTranslatedAttributes());
		}

		translation.setTranslatedEClasses(classes);
		translation.setTranslatedEReferences(references);
		translation.setTranslatedEAttributes(attributes);
	}
}
