/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.analysis;

import java.util.Collection;
import java.util.Map;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.Context;
import com.microsoft.z3.EnumSort;
import com.microsoft.z3.Expr;
import com.microsoft.z3.FuncDecl;
import com.microsoft.z3.Solver;

/**
 * Interface for generic translation of model product-lines to SMT.
 * 
 * @author bayha
 */
public interface IProductLineTranslation {

	/**
	 * Retrieves the context used for this translation.
	 * 
	 * @return The Z3 {@link Context}.
	 */
	public Context getContext();

	/**
	 * Translates the given model to Z3 SMT.
	 * 
	 * @param model
	 *            {@link EObject} with model to be translated.
	 */
	public void translateModel(EObject model);

	/**
	 * Translates the given product-line constraints to Z3 SMT.
	 * 
	 * @param constraints
	 *            A {@link Collection} of {@link IProductLineConstraint} to be translated.
	 * 
	 * @return A {@link Map} with Z3 {@link BoolExpr} to track the constraints in Z3
	 *         models.
	 */
	public Map<BoolExpr, IProductLineConstraint>
			translateConstraints(Collection<IProductLineConstraint> constraints);

	/**
	 * Retrieves the model element which was the source for the given Z3 expression.
	 * 
	 * @param expr
	 *            The Z3 {@link Expr} for which to retrieve the originating {@link EObject}.
	 * 
	 * @return The {@link EObject} which was the source for the given Z3 expression.
	 */
	public EObject getEObjectForExpression(Expr<?> expr);

	/**
	 * Creates and initializes the Z3 solver to be used for analyzes.
	 * 
	 * Note, that {@code translateModel()} and {@code translateConstraints} should be called prior
	 * to this method.
	 * 
	 * @return The Z3 {@link Solver} containing the model translation.
	 */
	public Solver createSolver();

	/**
	 * Retrieves the Z3 expression representing null for the given {@link EClass}.
	 * 
	 * @param cls
	 *            The {@link EClass} to retrieve the Z3 null element for.
	 * @return The Z3 {@link Expr} encoding null.
	 */
	public Expr<?> getNullElement(EClass cls);

	/**
	 * Retrieves the Z3 sort which represents the given EClass and its objects.
	 * 
	 * @param eClass
	 *            The {@link EClass} to get the Z3 sort for.
	 * @return The Z3 {@link EnumSort} for the given eClass.
	 */
	public EnumSort<?> getDatatype(EClass eClass);

	/**
	 * Retrieve the Z3 function declaration which represents the given attribute or reference for
	 * the given EClass.
	 * 
	 * @param eClass
	 *            The {@link EClass} for which to get a attribute or reference function.
	 * @param attRefName
	 *            {@link String} with the name for the attribute or reference.
	 * @return A Z3 {@link FuncDecl} representing the attribute or reference.
	 */
	FuncDecl<?> getAttributeReferenceFunctionDeclaration(EClass eClass, String attRefName);

	/**
	 * Specifies the classes for which model elements shall be translated.
	 * 
	 * Note, that all classes which are relevant for evaluating the constraints need to specified
	 * here, before staring a translation.
	 * 
	 * @param clss
	 *            A {@link Collection} of all {@link EClass} to be translated.
	 */
	public void setTranslatedEClasses(Collection<EClass> clss);

	/**
	 * Specifies the references which shall be translated.
	 * 
	 * Note, that all references which are used in the constraints need to specified
	 * here, before staring a translation.
	 * 
	 * @param refs
	 *            A {@link Collection} of all {@link EReference} to be translated.
	 */
	public void setTranslatedEReferences(Collection<EReference> refs);

	/**
	 * Specifies the attributes which shall be translated.
	 * 
	 * Note, that all attributes which are used in the constraints need to specified
	 * here, before staring a translation.
	 * 
	 * @param atts
	 *            A {@link Collection} of all {@link EAttribute} to be translated.
	 */
	public void setTranslatedEAttributes(Collection<EAttribute> atts);
}
