/*--------------------------------------------------------------------------+
|                                                                          |
| Copyright 2022 fortiss GmbH						                       |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.presence.compiler;

import static org.fortiss.variability.util.VariabilityUtilsInternal.getChildrenWithType;

import java.util.ArrayList;
import java.util.List;

import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.CharStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.variability.model.presence.LiteralPC;
import org.fortiss.variability.model.presence.PresenceCondition;
import org.fortiss.variability.presence.parser.PresenceConditionLexer;
import org.fortiss.variability.presence.parser.PresenceConditionParser;

/**
 * Compiler for {@link PresenceCondition}s.
 * 
 * @author bayha
 */
public class PresenceConditionCompiler {

	/** List of error messages from the last compiler run. */
	public List<String> errors = new ArrayList<String>();

	/**
	 * Parses a {@link String} to a {@link PresenceCondition}.
	 * 
	 * @param input
	 *            The {@link String}, representing a FeatureRestrictioun.
	 * @param context
	 *            Any element from the same containment tree, as the elements to
	 *            which the literals shall point.
	 * 
	 * @return A {@link PresenceCondition}, representing the 'input'.
	 */
	public PresenceCondition compilePresenceCondition(EObject context, String input) {
		errors.clear();

		if(input.trim().length() == 0) {
			return null;
		}

		PresenceConditionParser parser = null;
		CharStream in = new ANTLRInputStream(input);
		PresenceConditionLexer lexer = new PresenceConditionLexer(in);

		PresenceConditionErrorListener errorListener = new PresenceConditionErrorListener();

		lexer.addErrorListener(errorListener);

		CommonTokenStream tokenStream = new CommonTokenStream(lexer);
		parser = new PresenceConditionParser(tokenStream, context);

		parser.addErrorListener(errorListener);

		PresenceCondition result = parser.presenceConditionEOF().rest;

		errors.addAll(errorListener.getErrors());

		if(result != null) {
			EList<LiteralPC> literals = getChildrenWithType(result, LiteralPC.class);
			if(result instanceof LiteralPC) {
				literals.add((LiteralPC)result);
			}
			for(LiteralPC l : literals) {
				if(l.getLiteralReference() == null && l.getStringRepresentation() != null &&
						l.getStringRepresentation().length() > 0) {
					errors.add("The literal \"" + l.getStringRepresentation() +
							"\" does not denote any feature or element in this project.");
				}
			}
		}

		return result;
	}
}
